import logging

import coloredlogs


def configure():
    coloredlogs.install(
        level=logging.DEBUG,
        fmt="%(asctime)s %(name)s[%(process)d] %(levelname)s %(message)s",
    )
    logging.basicConfig()
    logging.root.setLevel(logging.DEBUG)
    logging.getLogger("websockets").setLevel(logging.INFO)
    logging.getLogger("asyncio").setLevel(logging.INFO)



import inspect
from typing import Dict, List

ROLES_MAP = {
    "bot": "assistant",
    "human": "user",
}


def convert_mml_to_llm_format(mml: List[Dict]) -> List[Dict]:
    """
    Converts the MML (Message Markup Language) format to the common LLM message format.
    Analogous to format_msgs_chain_to_llm_context in back/back/apps/language_model/consumers/__init__.py


    :param mml: List of messages in MML format
    :return: List of messages in LLM format {'role': 'user', 'content': '...'}
    """
    aggregated_messages = []
    current_role = None  # "user" for human and "assistant" for bot
    aggregated_contents = []  # list of Content objects for the current group

    def process_stack(stack: Dict) -> List[Dict]:
        """
        Process a single stack item into a list of LLM message format.
        """
        contents = []
        type = stack.get("type")

        if type == "message" or type == "message_chunk":
            contents.append(
                {
                    "type": "text",
                    "text": stack["payload"]["content"],
                }
            )
        elif type == "tool_use":
            contents.append(
                {
                    "type": "tool_use",
                    "tool_use": stack["payload"],
                }
            )
        elif type == "tool_result":
            contents.append(
                {
                    "type": "tool_result",
                    "tool_result": stack["payload"],
                }
            )

        return contents

    def process_msg(msg: Dict) -> List[Dict]:
        """
        Process each broker message into a list of LLM message format by iterating over its stacks.
        """
        contents = []
        for stack in msg.get("stack", []):
            contents.extend(process_stack(stack))
        return contents

    def merge_contents(existing: List[Dict], new: List[Dict]) -> List[Dict]:
        """
        Merge two lists of LLM message format.
        If the last element of the existing list and the first element of the new list are both text,
        then they are concatenated.
        """
        if not existing:
            return new
        if not new:
            return existing

        merged = existing.copy()
        if merged and new and merged[-1]["type"] == "text" and new[0]["type"] == "text":
            merged[-1]["text"] = (
                merged[-1]["text"].strip() + " " + new[0]["text"].strip()
            )
            merged.extend(new[1:])
        else:
            merged.extend(new)
        return merged

    for msg in mml:
        role = ROLES_MAP[msg["sender"]["type"]]
        msg_contents = process_msg(msg)
        if not msg_contents:
            continue

        if current_role is None:
            current_role = role
            aggregated_contents = msg_contents
        elif current_role == role:
            aggregated_contents = merge_contents(aggregated_contents, msg_contents)
        else:
            aggregated_messages.append(
                {
                    "role": current_role,
                    "content": aggregated_contents,
                }
            )
            current_role = role
            aggregated_contents = msg_contents

    if aggregated_contents:
        aggregated_messages.append(
            {
                "role": current_role,
                "content": aggregated_contents,
            }
        )

    return aggregated_messages


def function_to_json(func) -> dict:
    """
    Converts a Python function into a JSON-serializable dictionary
    that describes the function's signature, including its name,
    description, and parameters.
    Function from https://github.com/openai/swarm

    Args:
        func: The function to be converted.

    Returns:
        A dictionary representing the function's signature in JSON format.
    """
    type_map = {
        str: "string",
        int: "integer",
        float: "number",
        bool: "boolean",
        list: "array",
        dict: "object",
        type(None): "null",
    }

    try:
        signature = inspect.signature(func)
    except ValueError as e:
        raise ValueError(
            f"Failed to get signature for function {func.__name__}: {str(e)}"
        )

    parameters = {}
    for param in signature.parameters.values():
        try:
            param_type = type_map.get(param.annotation, "string")
        except KeyError as e:
            raise KeyError(
                f"Unknown type annotation {param.annotation} for parameter {param.name}: {str(e)}"
            )
        parameters[param.name] = {"type": param_type}

    required = [
        param.name
        for param in signature.parameters.values()
        if param.default == inspect._empty
    ]

    return {
        "type": "function",
        "function": {
            "name": func.__name__,
            "description": func.__doc__ or "",
            "parameters": {
                "type": "object",
                "properties": parameters,
                "required": required,
            },
        },
    }



from logging import getLogger
from typing import List
from uuid import uuid4

logger = getLogger(__name__)


class Layer:
    """
    Representation of all the future stack's layers. Implementing a new layer should inherit form this
    """

    _type = None
    _streaming = False

    def __init__(self, state={}):
        self.state = state
        self.stack_id = str(uuid4())

    async def build_payloads(self, ctx, data) -> tuple[List[dict], bool]:
        """
        Used to represent the layer as a dictionary which will be sent through the WS to the ChatFAQ's back-end server
        It is cached since there are layers as such as the LMGeneratedText which are computationally expensive
        :return:
            dict
                A json compatible dict
            bool
                If it is the last stack's layer or there are more stacks
        """
        raise NotImplementedError

    async def result(self, ctx, data, fsm_def_name: str = None) -> List[dict]:
        repr_gen = self.build_payloads(ctx, data)
        async for _repr, last_chunk in repr_gen:
            for r in _repr:
                r["type"] = self._type
                r["streaming"] = self._streaming
                r["meta"] = {}
                r["state"] = self.state
                if fsm_def_name:
                    r["fsm_definition"] = fsm_def_name
            yield [_repr, self.stack_id, last_chunk]


class Message(Layer):
    """
    A flexible message layer that can include text, references, tool calls, and other metadata.
    """

    _type = "message"

    def __init__(
        self,
        content,
        references={},
        tool_calls=[],
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.content = content
        self.references = references
        self.tool_calls = tool_calls

    async def build_payloads(self, ctx, data):
        payload = {
            "payload": {
                "content": self.content,
                "references": self.references,
                "tool_calls": self.tool_calls,
            }
        }
        yield [payload], True


class FileUpload(Layer):
    """
    A message layer that includes a file upload request.
    """
    _type = "file_upload"

    def __init__(
        self,
        content,
        file_extensions=[],
        max_size=0,
        *args,
        **kwargs,
    ):
        """
        :param file_extensions: A list of file extensions to request. For example: ["pdf", "xml"]
        :param max_size: The maximum size of the file to request in bytes. For example: 50 * 1024 * 1024 (50MB)
        """
        super().__init__(*args, **kwargs)
        self.content = content
        self.file_extensions = file_extensions
        self.max_size = max_size

    async def build_payloads(self, ctx, data):
        _payload = {
            "content": self.content,
            "files": {
                file_extension: {
                    "max_size": self.max_size,
                }
                for file_extension in self.file_extensions
            },
        }

        yield [{"payload": _payload}], True


class FileDownload(Layer):
    """
    A message layer that includes a file download.
    """
    _type = "file_download"

    def __init__(
            self,
            content: str,
            file_name: str,
            file_url: str,
            *args,
            **kwargs,
    ):
        """
        :param file_name: The name of the file. For example: "report.pdf"
        :param file_url: The URL of the file where the user can download it or visualize it. For example: "https://example.com/report.pdf"
        """
        super().__init__(*args, **kwargs)
        self.content = content
        self.file_name = file_name
        self.file_url = file_url

    async def build_payloads(self, ctx, data):
        payload = {
            "payload": {
                    "content": self.content,
                    "name": self.file_name,
                    "url": self.file_url,
            }
        }
        yield [payload], True


class StreamingMessage(Layer):
    _type = "message_chunk"
    _streaming = True

    def __init__(
        self,
        generator,
        references={},
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.generator = generator
        self.references = references

    async def build_payloads(self, ctx, data):
        async for chunk in self.generator:
            last_chunk = chunk.get("last_chunk", False)
            tool_calls = chunk.get("tool_calls", [])
            if last_chunk:  # now we send the references only in the final message
                payload = {
                    "payload": {
                        "content": chunk.get("content"),
                        "references": self.references,
                        "tool_calls": tool_calls,
                    }
                }
                yield (
                    [payload],
                    last_chunk,
                )
                break

            else:
                yield (
                    [
                        {
                            "payload": {
                                "content": chunk.get("content"),
                                "tool_calls": tool_calls,
                            }
                        }
                    ],
                    last_chunk,
                )


class GTMTag(Layer):
    _type = "gtm_tag"

    def __init__(self, tag):
        super().__init__()
        self.tag = tag

    async def build_payloads(self, ctx, data):
        payload = {"payload": self.tag}
        yield [payload], True


class UserFeedback(Layer):
    """
    An abstract class to inherit from for feedback layers representing common feedback elements.
    """

    def __init__(
        self,
        hint: str = None,
        merge_to_prev: bool = False,
        full_conversation: bool = False,
        *args,
        **kwargs,
    ):
        """
        :param hint: A text to explain the rating. For example: "Please rate the service"
        :param merge_to_prev: It controls whether the feedback should be merged (UI wise) with the previous message (default: False)
        :param full_conversation: Feedback reference always the last not feedback message from the bot, but they could reference to the whole conversation (default: False)
        """
        super().__init__(*args, **kwargs)
        self.hint = hint
        self.merge_to_prev = merge_to_prev
        self.full_conversation = full_conversation

    def _build_payloads(self):
        return {
            "hint": self.hint,
            "merge_to_prev": self.merge_to_prev,
            "full_conversation": self.full_conversation,
        }


class StarRating(UserFeedback):
    """
    A message layer that includes a star rating.
    """
    _type = "star_rating"

    def __init__(
        self,
        num_stars: int,
        explanation: str,
        *args,
        **kwargs,
    ):
        """
        :param num_stars: The maximum number of stars. For example: 5
        """
        super().__init__(*args, **kwargs, merge_to_prev=False)
        self.explanation = explanation
        self.num_stars = num_stars

    async def build_payloads(self, ctx, data):
        payload = {
            "payload": {
                **self._build_payloads(),
                "num_stars": self.num_stars,
                "explanation": self.explanation,
            }
        }
        yield [payload], True


class ThumbsRating(UserFeedback):
    """
    A message layer that includes a thumbs rating.
    """
    _type = "thumbs_rating"

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs, merge_to_prev=True)

    async def build_payloads(self, ctx, data):
        payload = {
            "payload": {
                **self._build_payloads(),
            }
        }
        yield [payload], True


class TextFeedback(UserFeedback):
    """
    A message layer that includes a feedback text box.
    """
    _type = "text_feedback"

    def __init__(
        self,
        placeholder: str,
        *args,
        **kwargs,
    ):
        """
        :param placeholder: The placeholder text inside the text box. For example: "Please provide your feedback here"
        """
        super().__init__(*args, **kwargs, merge_to_prev=False)
        self.placeholder = placeholder

    async def build_payloads(self, ctx, data):
        payload = {
            "payload": {
                **self._build_payloads(),
                "placeholder": self.placeholder,
            }
        }
        yield [payload], True


class CloseConversation(Layer):
    """
    A message layer that includes a close conversation request.
    """
    _type = "close_conversation"

    async def build_payloads(self, ctx, data):
        yield [{"payload": {}}], True


class ToolUse(Layer):
    """
    A message layer that includes a tool use request.
    """
    _type = "tool_use"

    def __init__(self, name: str = None, id: str = None, args: dict = None):
        super().__init__()
        self.id = id
        self.name = name
        self.args = args

    async def build_payloads(self, ctx, data):
        payload = {
            "payload": {
                "id": self.id,
                "name": self.name,
                "args": self.args,
            }
        }
        yield [payload], True

class ToolResult(Layer):
    """
    A message layer that includes a tool result.
    """
    _type = "tool_result"

    def __init__(self, id: str = None, name: str = None, result: dict = None):
        super().__init__()
        self.id = id
        self.name = name
        self.tool_result = result

    async def build_payloads(self, ctx, data):
        payload = {
            "payload": {
                "id": self.id,
                "name": self.name,
                "result": self.tool_result,
            }
        }
        yield [payload], True



import inspect
import os
import json
from enum import Enum
from logging import getLogger
from typing import List, Callable, Any

from chatfaq_sdk import ChatFAQSDK
from chatfaq_sdk.clients import query_prompt_default, llm_request
from chatfaq_sdk.layers import Message, ToolUse, ToolResult, StreamingMessage, Layer

logger = getLogger(__name__)


class MessageSender(Enum):
    system = "system"
    assistant = "assistant"
    user = "user"

class MessageSenderEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, MessageSender):
            return obj.value
        return super().default(obj)

class StreamingMessageWithReferences(Layer):
    """
    This layer is used to send a streaming message with references to the user.
    The special thing is that the references come in the last chunk, instead of when initializing the layer as done in StreamingMessage
    """
    _type = "message_chunk"
    _streaming = True

    def __init__(
        self,
        generator,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.generator = generator

    async def build_payloads(self, ctx, data):
        async for chunk in self.generator:
            references = chunk.get("references", {})
            if references:  # now we send the references only in the final message
                payload = {
                    "payload": {
                        "content": chunk.get("content"),
                        "references": references,
                        "tool_calls": [],
                    }
                }
                yield (
                    [payload],
                    True,  # last_chunk
                )

            else:
                yield (
                    [
                        {
                            "payload": {
                                "content": chunk.get("content"),
                                "tool_calls": [],
                            }
                        }
                    ],
                    False,  # last_chunk
                )


class AgentAbs:
    intro_msg = ""

    def __init__(self):
        self.conversation = []
        self.prompt = None
        self.add_assistant_message(self.intro_msg)

    def add_message(self, sender: MessageSender, message: str):
        self.conversation.append((sender, message))

    def add_user_message(self, message: Any):
        self.add_message(MessageSender.user, message)

    def add_assistant_message(self, message: str):
        self.add_message(MessageSender.assistant, message)

    def _from_serialized(self, data):
        pass

    def _serialize(self):
        return {}

    async def _async_init(self, sdk: ChatFAQSDK, *args, **kwargs):
        raise NotImplementedError

    @classmethod
    def from_serialized(cls, data):
        instance = cls()
        instance.conversation = []

        for s, m in data['conversation']:
            instance.add_message(MessageSender(s), m)

        instance._from_serialized(data)

        return instance

    def serialize(self):
        return {
            "conversation": [(s.value, m) for s, m in self.conversation],
            **self._serialize()
        }

    @classmethod
    async def async_init(cls, sdk: ChatFAQSDK, *args, **kwargs):
        instance = cls()
        await instance._async_init(sdk, *args, **kwargs)
        return instance

    def format_conversation(self):
        res = []
        for sender, message in self.conversation:
            res.append(
                {
                    "role": sender.value,
                    "content": message,
                }
            )
        return [
            {
                "role": MessageSender.system.value if self.conversation else MessageSender.user.value,
                "content": self.prompt,
            },
            *res,
        ]

    async def set_prompt(
        self, sdk: ChatFAQSDK, prompt_name: str, default_prompt: str
    ):
        self.prompt = await sdk.query_prompt_default(prompt_name, default_prompt)

    async def tool_use_loop(self, sdk: ChatFAQSDK, llm: str, ctx: dict, tools: List[Callable], logging=False):
        if logging:
            logger.info("\n" + "-" * 50 + "      TOOL USE LOOP \n")
            logger.info("\033[42m" + "\033[30m tools \033[0m")
            logger.info("\033[92m" + str(tools) + "\033[0m")

        while True:
            messages = self.format_conversation()
            if logging:
                logger.info("\n" + "-" * 50 + "      PROMPT \n")
                logger.info("\033[43m" + "\033[30m prompt \033[0m")
                logger.info("\033[93m" + messages[0]["content"] + "\033[0m")

            response = await llm_request(
                sdk,
                llm,
                use_conversation_context=False,
                conversation_id=ctx["conversation_id"],
                bot_channel_name=ctx["bot_channel_name"],
                messages=messages,
                tools=tools,
                tool_choice="auto",
                stream=False,
            )
            if logging:
                logger.info("\n" + "-" * 50 + "      RESPONSE \n")
                logger.info("\033[45m" + "\033[30m response \033[0m")
                logger.info("\033[95m" + str(response) + "\033[0m")

            tool_results = []
            for content in response["content"]:
                if content["type"] == "text":
                    self.add_assistant_message(content["text"])
                    yield Message(content["text"])
                elif content["type"] == "tool_use":
                    tool_use = content["tool_use"]
                    # Find the corresponding tool
                    tool = None
                    for t in tools:
                        if t.__name__ == tool_use["name"]:
                            tool = t
                    if not tool:
                        raise ValueError(f"Tool {tool_use['name']} not found")
                    # Execute the tool
                    try:
                        if inspect.iscoroutinefunction(tool):
                            result = await tool(**tool_use["args"], sdk=sdk, ctx=ctx, agent=self)
                        else:
                            result = tool(**tool_use["args"], sdk=sdk, ctx=ctx, agent=self)
                    except Exception as e:
                        result = f"Error executing tool {tool_use['name']}: {str(e)}"

                    if inspect.isasyncgen(result):
                        yield StreamingMessageWithReferences(result)
                        result = str("submitted")

                    yield ToolUse(name=tool_use["name"], id=tool_use["id"], args=tool_use["args"])
                    yield ToolResult(
                        id=tool_use["id"], name=tool_use["name"], result=result
                    )
                    tool_results.append({
                        "id": tool_use["id"],
                        "name": tool_use["name"],
                        "result": result
                    })
            if not tool_results:
                break
            # Append assistant and user messages for the next iteration
            self.add_assistant_message(response["content"])
            self.add_user_message([
                {"type": "tool_result", "tool_result": tr} for tr in tool_results
            ])

    async def conversation_loop(self, sdk, ctx):
        raise NotImplementedError

    def solved(self):
        raise NotImplementedError



from logging import getLogger
from typing import Callable, Dict, List, Optional, Union

from chatfaq_sdk import ChatFAQSDK
from chatfaq_sdk.types import CacheConfig
from chatfaq_sdk.utils import function_to_json

logger = getLogger(__name__)


def llm_request(
    sdk: ChatFAQSDK,
    llm_config_name: str,
    messages: List[Dict] = None,
    temperature: float = 0.7,
    max_tokens: int = 4096,
    seed: int = 42,
    thinking: str | Dict = None,
    tools: List[Union[Callable, Dict]] = None,
    tool_choice: str = None,
    conversation_id: str = None,
    bot_channel_name: str = None,
    use_conversation_context: bool = True,
    cache_config: Optional[CacheConfig] = None,
    response_schema: Optional[Dict] = None,
    stream: bool = False,
):
    """
    Dispatches an LLM request either in streaming or non-streaming mode.

    When `stream` is True, returns an async generator so you can iterate over
    result chunks using `async for ... in llm_request(...)`.

    When `stream` is False, returns an awaitable coroutine that resolves to
    the final (non‑streamed) result, so you would use:

      response = await llm_request(..., stream=False)
    """
    if stream:
        return _llm_request_streaming(
            sdk,
            llm_config_name,
            messages,
            temperature,
            max_tokens,
            seed,
            thinking,
            tools,
            tool_choice,
            conversation_id,
            bot_channel_name,
            use_conversation_context,
            cache_config,
            response_schema,
            stream,
        )
    else:
        return _llm_request_non_streaming(
            sdk,
            llm_config_name,
            messages,
            temperature,
            max_tokens,
            seed,
            thinking,
            tools,
            tool_choice,
            conversation_id,
            bot_channel_name,
            use_conversation_context,
            cache_config,
            response_schema,
            stream,
        )


async def _llm_request_non_streaming(
    sdk: ChatFAQSDK,
    llm_config_name: str,
    messages: List[Dict] = None,
    temperature: float = 0.7,
    max_tokens: int = 4096,
    seed: int = 42,
    thinking: str | Dict = None,
    tools: List[Union[Callable, Dict]] = None,
    tool_choice: str = None,
    conversation_id: str = None,
    bot_channel_name: str = None,
    use_conversation_context: bool = True,
    cache_config: Optional[CacheConfig] = None,
    response_schema: Optional[Dict] = None,
    stream: bool = False,  # always False in this branch
):
    if tools:
        tools = [function_to_json(tool) if isinstance(tool, Callable) else tool for tool in tools]

    await sdk.send_llm_request(
        llm_config_name,
        messages,
        temperature,
        max_tokens,
        seed,
        thinking,
        tools,
        tool_choice,
        conversation_id,
        bot_channel_name,
        use_conversation_context,
        cache_config,
        response_schema,
        stream,
    )

    logger.debug("[LLMRequest] Waiting for non-streaming LLM req...")
    results = (await sdk.llm_request_futures[bot_channel_name])()
    logger.debug("[LLMRequest] Received non-streaming response from LLM req")
    # If a single chunk is received, unwrap it before returning.
    return results[0] if len(results) == 1 else results


async def _llm_request_streaming(
    sdk: ChatFAQSDK,
    llm_config_name: str,
    messages: List[Dict] = None,
    temperature: float = 0.7,
    max_tokens: int = 4096,
    seed: int = 42,
    thinking: str | Dict = None,
    tools: List[Union[Callable, Dict]] = None,
    tool_choice: str = None,
    conversation_id: str = None,
    bot_channel_name: str = None,
    use_conversation_context: bool = True,
    cache_config: Optional[CacheConfig] = None,
    response_schema: Optional[Dict] = None,
    stream: bool = True,  # always True in this branch
):
    if tools:
        tools = [function_to_json(tool) if isinstance(tool, Callable) else tool for tool in tools]

    await sdk.send_llm_request(
        llm_config_name,
        messages,
        temperature,
        max_tokens,
        seed,
        thinking,
        tools,
        tool_choice,
        conversation_id,
        bot_channel_name,
        use_conversation_context,
        cache_config,
        response_schema,
        stream,
    )

    logger.debug("[LLMRequest] Waiting for streaming LLM req...")
    final = False
    while not final:
        results = (await sdk.llm_request_futures[bot_channel_name])()
        logger.debug("[LLMRequest] ...received streaming results from LLM req")
        for result in results:
            final = result.get("last_chunk", False)
            yield result
        logger.debug(f"[LLMRequest] (Final: {final})")


async def retrieve(
    sdk: ChatFAQSDK,
    retriever_name: str,
    query: str,
    top_k: int,
    bot_channel_name: str,
):
    await sdk.send_retriever_request(retriever_name, query, top_k, bot_channel_name)

    logger.debug("[Retrieve] Waiting for Retrieve req...")
    results = await sdk.retriever_request_futures[bot_channel_name]
    logger.debug("[Retrieve] ...receive results from Retrieve req")

    return results


async def query_kis(
    sdk: ChatFAQSDK,
    knowledge_base_name: str,
    query: Optional[Dict] = None,
):
    return await sdk.query_kis(knowledge_base_name, query)


async def query_prompt(
    sdk: ChatFAQSDK,
    prompt_name: str,
):
    return await sdk.query_prompt(prompt_name)

async def query_prompt_default(
    sdk: ChatFAQSDK,
    prompt_name: str,
    default_prompt: str,
):
    return await sdk.query_prompt_default(prompt_name, default_prompt)

async def get_prompt(
    sdk: ChatFAQSDK,
    prompt_config_name: str,
    bot_channel_name: str,
):
    await sdk.send_prompt_request(prompt_config_name, bot_channel_name)

    logger.debug("[Prompt] Waiting for Prompt req...")
    results = await sdk.prompt_request_futures[bot_channel_name]
    logger.debug("[Prompt] ...receive results from Prompt req")
    return results



from enum import Enum


class MessageType(Enum):
    fsm_def = "fsm_def"
    rpc_request = "rpc_request"
    rpc_result = "rpc_result"
    llm_request = "llm_request"
    llm_request_result = "llm_request_result"
    retriever_request = "retriever_request"
    retriever_request_result = "retriever_request_result"
    prompt_request = "prompt_request"
    prompt_request_result = "prompt_request_result"
    error = "error"
    register_parsers = "register_parsers"
    parser_result_ki = "parser_result_ki"
    parser_finished = "parser_finished"


class RPCNodeType(Enum):
    action = "action"
    condition = "condition"



import asyncio
import copy
import inspect
import json
import os
import queue
import urllib.parse
import uuid
from functools import wraps
from logging import getLogger
from typing import Callable, List, Optional, Union

import httpx
import sentry_sdk
import websockets

from chatfaq_sdk import settings
from chatfaq_sdk.conditions import Condition
from chatfaq_sdk.data_source_parsers import DataSourceParser
from chatfaq_sdk.fsm import FSMDefinition
from chatfaq_sdk.layers import Layer
from chatfaq_sdk.types import DataSource, KnowledgeItem, WSType
from chatfaq_sdk.types.messages import MessageType, RPCNodeType

settings.configure()

logger = getLogger()


class ChatFAQSDK:
    """
    This SDK helps on:
        - Defining the FSM handlers for transitions & states
        - Declare the FSM in ChatFAQ's back-end server
        - Translate inbound messages from ChatFAQ's back-end server into function calls (handlers) and vice-versa
    """

    def __init__(
        self,
        chatfaq_ws: str,
        chatfaq_http: str,
        token: str,
        fsm_name: Optional[Union[int, str]],
        fsm_definition: Optional[FSMDefinition] = None,
        overwrite_definition: Optional[bool] = False,
        data_source_parsers: Optional[dict[str, DataSourceParser]] = None,
        authentication_required: bool = False
    ):
        """
        Parameters
        ----------
        chatfaq_ws: str
            The WS address of your ChatFAQ's back-end server

        token: str
            The auth token of your ChatFAQ's back-end server

        fsm_name: Union[int, str, None]
            The id or name of the FSM you are going to associate this client to. If you are going to create a new FSM
            then it should be the name you are going to give to the new created FSM

        fsm_definition: Union[FSMDefinition, None]
            The FSM you are going to create in the ChatFAQ's back-end server, if already exists a FSM definition on the server with
            the same struincurre then that one will be reused and your 'name' parameter will be ignored

        overwrite_definition: Optional[bool]
            Whether to overwrite an existing FSM with the same name. Watch out! if you set this to True and there is a FSM
            with the same name it will be deleted and recreated with the new definition, if there are other instances of the FSM running
            they will get into a undefined state use it carefully only when you are sure that no other instance is running

        data_source_parsers: Optional[dict[str, DataSourceParser]]
            A dictionary with the parsers you want to register in the ChatFAQ's back-end server, the key should be the name of the parser
            and the value should be the parser function itself
        """
        if fsm_definition is dict and fsm_name is None:
            raise Exception("If you declare a FSM definition you should provide a name")
        self.chatfaq_ws = chatfaq_ws
        self.chatfaq_http = chatfaq_http
        self.token = token
        self.fsm_name = fsm_name
        self.fsm_def = fsm_definition
        self.overwrite_definition = overwrite_definition
        self.data_source_parsers = data_source_parsers
        self.authentication_required = authentication_required
        self.rpcs = {}
        # _rpcs is just an auxiliary variable to register the rpcs without the decorator function just so we know if we
        # already registered that rpc under that name and avoid duplicates
        self._rpcs = {}

        self.llm_request_futures = {}
        self.llm_request_msg_buffer = {}
        self.retriever_request_futures = {}
        self.prompt_request_futures = {}

        if self.fsm_def is not None:
            self.fsm_def.register_rpcs(self)

        self.sentry = None
        if SENTRY_DSN := os.getenv("SENTRY_DSN"):
            self.sentry = sentry_sdk.init(SENTRY_DSN)

    def connect(self):
        try:
            asyncio.run(self.connexions())
        except KeyboardInterrupt:
            asyncio.run(self._disconnect())
        except Exception as e:
            if self.sentry:
                sentry_sdk.capture_exception(e)
            raise e

    async def connexions(self):
        setattr(self, f"ws_{WSType.rpc.value}", None)
        setattr(self, f"ws_{WSType.ai.value}", None)
        rpc_actions = {
            MessageType.rpc_request.value: self.rpc_request_callback,
            MessageType.error.value: self.error_callback,
        }
        ai_actions = {
            MessageType.llm_request_result.value: self.llm_request_result_callback,
            MessageType.retriever_request_result.value: self.retriever_request_result_callback,
            MessageType.prompt_request_result.value: self.prompt_request_result_callback,
            MessageType.error.value: self.error_callback,
        }
        coros_or_futures = [
            self.consumer(WSType.rpc.value, on_connect=self.on_connect_rpc),
            self.consumer(WSType.ai.value, on_connect=None),
            self.producer(rpc_actions, WSType.rpc.value),
            self.producer(ai_actions, WSType.ai.value),
        ]
        if self.data_source_parsers:
            setattr(self, f"ws_{WSType.parse.value}", None)
            parser_actions = {
                key: self.parsing_wrapper(value)
                for key, value in self.data_source_parsers.items()
            }
            parser_actions[MessageType.error.value] = self.error_callback
            coros_or_futures += [
                self.consumer(WSType.parse.value, on_connect=self.on_connect_parsing),
                self.producer(parser_actions, WSType.parse.value),
            ]

        await asyncio.gather(
            *coros_or_futures,
        )

    async def consumer(self, consumer_route, on_connect=None):
        setattr(self, f"queue_{consumer_route}", queue.Queue())
        uri = urllib.parse.urljoin(self.chatfaq_ws, f"back/ws/broker/{consumer_route}/")
        if (
            consumer_route == WSType.rpc.value
            and self.fsm_name is not None
            and self.fsm_def is None
        ):
            uri = f"{uri}{self.fsm_name}/"

        parsed_token = urllib.parse.quote(self.token)
        uri = f"{uri}?token={parsed_token}"
        while True:
            try:
                logger.info(f"[{consumer_route.upper()}] Connecting to {uri}")
                async with websockets.connect(uri) as ws:
                    logger.info(f"[{consumer_route.upper()}] Connected")
                    setattr(self, f"ws_{consumer_route}", ws)
                    if on_connect is not None:
                        await on_connect()
                    logger.info(
                        f"[{consumer_route.upper()}] ---------------------- Listening..."
                    )
                    await self._consume_loop(
                        consumer_route
                    )  # <----- "infinite" Connection Loop
            except (websockets.WebSocketException, ConnectionRefusedError) as e:
                logger.info(f"{consumer_route.upper()} Connection error: {e}, retrying...")
                await asyncio.sleep(1)

    async def _consume_loop(self, consumer_route):
        while True:
            data = json.loads(await getattr(self, f"ws_{consumer_route}").recv())
            getattr(self, f"queue_{consumer_route}").put(data)

    async def producer(self, actions, consumer_route):
        ws_attrs = [attr for attr in dir(self) if attr.startswith("ws_")]
        while True:
            if any(
                getattr(self, ws_attr) is None or not getattr(self, ws_attr).open
                for ws_attr in ws_attrs
            ):
                await asyncio.sleep(0.01)
                continue
            try:
                data = getattr(self, f"queue_{consumer_route}").get(False)
            except queue.Empty:
                await asyncio.sleep(0.01)
                continue

            if actions.get(data.get("type")) is not None:
                asyncio.create_task(actions[data.get("type")](data["payload"]))
            else:
                logger.error(f"Unknown action type: {data.get('type')}")


    async def retriever_request_result_callback(self, payload):
        logger.info(f"[RETRIEVER] Result received: {payload}")
        self.retriever_request_futures[payload["bot_channel_name"]].set_result(payload)


    async def prompt_request_result_callback(self, payload):
        logger.info(f"[PROMPT] Result received: {payload}")
        self.prompt_request_futures[payload["bot_channel_name"]].set_result(payload)

    async def on_connect_rpc(self):
        if self.fsm_def is not None:
            logger.info(f"[RPC] Setting FSM by definition {self.fsm_name}" + ", overwriting existing definition if already exists." if self.overwrite_definition else "")
            await getattr(self, f"ws_{WSType.rpc.value}").send(
                json.dumps(
                    {
                        "type": MessageType.fsm_def.value,
                        "data": {
                            "name": self.fsm_name,
                            "definition": self.fsm_def.to_dict_repr(),
                            "overwrite": self.overwrite_definition,
                            "authentication_required": self.authentication_required,
                        },
                    }
                )
            )

    async def on_connect_parsing(self):
        if self.data_source_parsers is not None:
            parsers = list(self.data_source_parsers.keys())
            logger.info(f"[PARSE] Registering Data Source Parsers {parsers}")
            await getattr(self, f"ws_{WSType.parse.value}").send(
                json.dumps(
                    {
                        "type": MessageType.register_parsers.value,
                        "data": {
                            "parsers": parsers,
                        },
                    }
                )
            )

    async def _disconnect(self):
        logger.info("Shutting Down...")
        wss = [getattr(self, attr) for attr in dir(self) if attr.startswith("ws_")]
        for ws in wss:
            if ws is not None and ws.open:
                await ws.close()

    async def rpc_request_callback(self, payload):
        logger.info(f"[RPC] Executing ::: {payload['name']}")
        status = {}
        if self.fsm_def.status_class:
            status = await self.fsm_def.status_class.deserialize(self, payload["ctx"], payload["ctx"]["status"])
        payload["ctx"]["status"] = status

        for index, state_or_transition in enumerate(self.rpcs[payload["name"]]):
            stack_group_id = str(uuid.uuid4())
            logger.info(f"[RPC]     |---> ::: {state_or_transition}")

            async for res, stack_id, last_chunk, node_type, last_from_state_or_transition in self._run_state_or_transition(
                state_or_transition, payload["ctx"]
            ):
                _ctx = {**payload["ctx"]}
                if _ctx["status"]:
                    _ctx["status"] = _ctx["status"].serialize()

                await getattr(self, f"ws_{WSType.rpc.value}").send(
                    json.dumps(
                        {
                            "type": MessageType.rpc_result.value,
                            "data": {
                                "ctx": _ctx,
                                "node_type": node_type,
                                "stack_group_id": stack_group_id,
                                "stack_id": stack_id,
                                "stack": res,
                                "last_chunk": last_chunk,
                                "last": index
                                == len(self.rpcs[payload["name"]]) - 1
                                and last_from_state_or_transition,
                            },
                        }
                    )
                )

    async def llm_request_result_callback(self, payload):
        # mesages could come at a faster rate than the handler can process them, so we need to buffer them
        if self.llm_request_msg_buffer.get(payload["bot_channel_name"]) is None:
            self.llm_request_msg_buffer[payload["bot_channel_name"]] = []
        self.llm_request_msg_buffer[payload["bot_channel_name"]].append(payload)

        # then we set future result to the generator as an indicator to the handler that it can start processing the
        # messages. The generator will be consumed by the handler once awaited, and it will take care of resetting the
        # buffer and setting the future again, in the meanwhile messages can still arrive and we keep buffering
        if not self.llm_request_futures[payload["bot_channel_name"]].done():
            self.llm_request_futures[payload["bot_channel_name"]].set_result(
                self.llm_result_streaming_generator(payload["bot_channel_name"])
            )

    def llm_result_streaming_generator(self, bot_channel_name):
        # The generator will be consumed by the handler once awaited, and it will take care of resetting the buffer and
        # setting the future again
        def _llm_result_streaming_generator():
            self.llm_request_futures[bot_channel_name] = (
                asyncio.get_event_loop().create_future()
            )

            _message_buffer = copy.deepcopy(
                self.llm_request_msg_buffer[bot_channel_name]
            )

            self.llm_request_msg_buffer[bot_channel_name] = []

            return _message_buffer

        return _llm_result_streaming_generator

    @staticmethod
    async def error_callback(payload):
        logger.error(f"Error from ChatFAQ's back-end server: {payload}")

    async def send_llm_request(
        self,
        llm_config_name,
        messages,
        temperature,
        max_tokens,
        seed,
        thinking,
        tools,
        tool_choice,
        conversation_id,
        bot_channel_name,
        use_conversation_context,
        cache_config,
        response_schema,
        stream,
    ):
        logger.info(f"[LLM] Requesting LLM ({llm_config_name})")
        self.llm_request_futures[bot_channel_name] = (
            asyncio.get_event_loop().create_future()
        )
        await getattr(self, f"ws_{WSType.ai.value}").send(
            json.dumps(
                {
                    "type": MessageType.llm_request.value,
                    "data": {
                        "llm_config_name": llm_config_name,
                        "messages": messages,
                        "conversation_id": conversation_id,
                        "bot_channel_name": bot_channel_name,
                        "temperature": temperature,
                        "max_tokens": max_tokens,
                        "seed": seed,
                        "thinking": thinking,
                        "tools": tools,
                        "tool_choice": tool_choice,
                        "use_conversation_context": use_conversation_context,
                        "cache_config": cache_config.__dict__ if cache_config else None,
                        "response_schema": response_schema,
                        "stream": stream,
                    },
                }
            )
        )

    async def send_retriever_request(
        self, retriever_config_name, query, top_k, bot_channel_name
    ):
        logger.info(f"[RETRIEVER] Requesting Retriever ({retriever_config_name})")
        self.retriever_request_futures[bot_channel_name] = (
            asyncio.get_event_loop().create_future()
        )
        await getattr(self, f"ws_{WSType.ai.value}").send(
            json.dumps(
                {
                    "type": MessageType.retriever_request.value,
                    "data": {
                        "retriever_config_name": retriever_config_name,
                        "bot_channel_name": bot_channel_name,
                        "query": query,
                        "top_k": top_k,
                    },
                }
            )
        )

    async def retriever_request(
        self, retriever_config_name, query, top_k=3
    ):
        async with httpx.AsyncClient() as client:
            response = await client.get(
                urllib.parse.urljoin(
                    self.chatfaq_http,
                    f"back/api/language-model/retrieve/?retriever_config__name={retriever_config_name}&query={query}&top_k={top_k}"
                ),
                headers={"Authorization": f"Token {self.token}"},
            )
            response.raise_for_status()
            return response.json()

    async def query_kis(self, knowledge_base_name, query: Optional[dict] = None) -> List[KnowledgeItem]:
        res = []
        offset = 0
        more = True
        while more:
            async with httpx.AsyncClient() as client:
                response = await client.get(
                    urllib.parse.urljoin(
                        self.chatfaq_http,
                        f"back/api/language-model/knowledge-items/?knowledge_base__name={knowledge_base_name}&metadata={json.dumps(query)}&offset={offset}",
                    ),
                    headers={"Authorization": f"Token {self.token}"},
                )
                response.raise_for_status()
                _results = response.json()["results"]
                _total = response.json()["count"]
                res += _results
                offset = len(res)
                more = offset < _total

        return [KnowledgeItem(**r) for r in res]

    async def query_prompt(self, prompt_name) -> List[KnowledgeItem]:
        async with httpx.AsyncClient() as client:
            response = await client.get(
                urllib.parse.urljoin(
                    self.chatfaq_http,
                    f"back/api/language-model/prompt-configs/?name={prompt_name}",
                ),
                headers={"Authorization": f"Token {self.token}"},
            )
            response.raise_for_status()
            results = response.json()["results"]
            if results:
                return results[0]["prompt"]

    async def query_prompt_default(self, prompt_name, default_prompt) -> List[KnowledgeItem]:
        """
        Queries the prompt with the given name, if not found returns the default prompt.
        """
        try:
            prompt = await self.query_prompt(prompt_name)
        except httpx.HTTPStatusError as e:
            logger.error(f"Error querying prompt {prompt_name}: {e}")
            return default_prompt

        if prompt is None:
            logger.warning(f"{prompt_name} not found, using default")
            return default_prompt
        else:
            logger.info(f"{prompt_name} found")
            return prompt

    async def submit_feedback_last_bot_msg(self, ctx: dict, value: str, comment: str):
        conv_mml = ctx.get("conv_mml", [])
        last_bot_message_id = None
        for i in reversed(conv_mml):
            if i["sender"]["type"] == "bot" and i["stack"]:
                last_bot_message_id = i["id"]
                break
        if last_bot_message_id:
            async with httpx.AsyncClient() as client:
                response = await client.post(
                    urllib.parse.urljoin(
                        self.chatfaq_http,
                        f"back/api/broker/user-feedback/",
                    ),
                    json={
                        "message_source": last_bot_message_id,
                        "message_target": last_bot_message_id,
                        "feedback_data": {
                            "thumb_value": value,
                            "feedback_comment": comment,
                        },
                    },
                    headers={"Authorization": f"Token {self.token}"},
                )
                response.raise_for_status()
                return response.json()
        return False

    async def send_prompt_request(self, prompt_config_name, bot_channel_name):
        logger.info(f"[PROMPT] Requesting Prompt ({prompt_config_name})")
        self.prompt_request_futures[bot_channel_name] = (
            asyncio.get_event_loop().create_future()
        )
        await getattr(self, f"ws_{WSType.ai.value}").send(
            json.dumps(
                {
                    "type": MessageType.prompt_request.value,
                    "data": {
                        "prompt_config_name": prompt_config_name,
                        "bot_channel_name": bot_channel_name,
                    },
                }
            )
        )

    def rpc(self, name: str) -> Callable:
        """
        Decorator for registering functions as handlers
        Parameters
        ----------
        name : str
            Name to which the function will be called once we received it from ChatFAQ's back-end server
        """

        def outer(func):
            @wraps(func)
            def inner(sdk: ChatFAQSDK, ctx: dict):
                return func(sdk, ctx)

            if name not in self.rpcs:
                self._rpcs[name] = []
                self.rpcs[name] = []
            if func not in self._rpcs[name]:
                self._rpcs[name].append(func)
                self.rpcs[name].append(inner)

            return inner

        return outer

    async def _run_state_or_transition(self, state_or_transition, data):
        _state_or_transition = state_or_transition(self, data)
        _last_result = None
        if inspect.isasyncgen(_state_or_transition):
            async for layer in _state_or_transition:
                async for results in self._layer_results(layer, data):
                    _last_result = results
                    yield [*results, False]
        else:
            layer = await _state_or_transition
            async for results in self._layer_results(layer, data):
                yield [*results, True]
                return
        yield [[], str(uuid.uuid4()), True, _last_result[-1], True]

    async def _layer_results(self, layer, data):
        if not isinstance(layer, Layer) and not isinstance(layer, Condition):
            raise Exception(
                "RPCs results should return either Layers type objects or result type objects"
            )
        results = layer.result(self, data, fsm_def_name=self.fsm_name)
        # check if is generator
        async for r in results:
            yield r + [
                RPCNodeType.action.value
                if isinstance(layer, Layer)
                else RPCNodeType.condition.value
            ]

    def parsing_wrapper(self, parser):
        async def _parsing_wrapper(payload):
            logger.info(f"[PARSE] Parsing ::: {payload}")
            data_source = DataSource(**payload)

            for ki in parser(data_source.kb_id, data_source.ds_id, data_source):
                async with httpx.AsyncClient() as client:
                    response = await client.post(
                        urllib.parse.urljoin(
                            self.chatfaq_http,
                            "back/api/language-model/knowledge-items/",
                        ),
                        json=ki.dict(),
                        headers={"Authorization": f"Token {self.token}"},
                    )
                    response.raise_for_status()
                    ki_res = response.json()
                    if ki.images:
                        for index, ki_image in enumerate(ki.images):
                            ki_image.knowledge_item = ki_res["id"]
                            response = await client.post(
                                urllib.parse.urljoin(
                                    self.chatfaq_http,
                                    "back/api/language-model/knowledge-item-images/",
                                ),
                                data=ki_image.dict(),
                                files=ki_image.files(),
                                headers={"Authorization": f"Token {self.token}"},
                            )
                            response.raise_for_status()
                            ki_image = response.json()
                            ki_res["content"] = ki_res["content"].replace(
                                f"[[Image {index}]]",
                                f"![{ki_image.get('image_caption') or ki_image['image_file_name']}]({ki_image['image_file_name']})",
                            )
                            response = await client.patch(
                                urllib.parse.urljoin(
                                    self.chatfaq_http,
                                    f"back/api/language-model/knowledge-items/{ki_res['id']}/",
                                ),
                                json=ki_res,
                                headers={"Authorization": f"Token {self.token}"},
                            )
                            response.raise_for_status()

            if data_source.task_id:
                await getattr(self, f"ws_{WSType.parse.value}").send(
                    json.dumps(
                        {
                            "type": MessageType.parser_finished.value,
                            "data": {"task_id": data_source.task_id},
                        }
                    )
                )

        return _parsing_wrapper



from typing import Iterator

from chatfaq_sdk.types import KnowledgeItem, DataSource


class DataSourceParser:
    def parse(self, data_source: DataSource) -> Iterator[KnowledgeItem]:
        raise NotImplementedError()

    def __call__(self, kb_id, ds_id, data_source: DataSource) -> Iterator[KnowledgeItem]:
        for ki in self.parse(data_source):
            ki.knowledge_base = kb_id
            ki.data_source = ds_id
            yield ki



class Condition:
    """
    Instances of this class should be always returned by the conditions of the FSM's transitions
    """

    def __init__(self, score: float, data: dict = {}):
        """

        Parameters
        ----------
        score: float
            The confidence of the transition, this scalar will be used by the FSM to determine at the end of the
            executiron of all the possible transitions which one should be the right one
        data: dict
            This data is passed from the transition result to the resulting state
            this is usually helpful when the transition compute something expensive, and we don't want to recompute
            again in the state
        """

        self.score = score
        self.data = data


    def __or__(self, other):
        return Condition(max(self.score, other.score), {**self.data, **other.data})

    def __and__(self, other):
        return Condition(min(self.score, other.score), {**self.data, **other.data})

    def __sub__(self, other):
        other_score = other.score if isinstance(other, Condition) else other
        return Condition(max(0, self.score - other_score), {**self.data, **other.data})

    def __add__(self, other):
        other_score = other.score if isinstance(other, Condition) else other
        return Condition(self.score + other_score, {**self.data, **other.data})

    def __eq__(self, other):
        if isinstance(other, Condition):
            return self.score == other.score and self.data == other.data
        return False

    def __repr__(self):
        return f"Condition(score={self.score}, data={self.data})"

    async def result(self, *args, fsm_def_name: str = None):
        yield [[{"score": self.score, "data": self.data}], 0, True]



from dataclasses import dataclass
from enum import Enum
from typing import Any, List, Optional


@dataclass
class KnowledgeItemImage:
    """
    An image contained in a KnowledgeItem.
    """

    image_name: str
    image_bytes: bytes
    knowledge_item: str = None
    image_caption: Optional[str] = None

    def dict(self):
        return {
            "knowledge_item": self.knowledge_item,
            "image_caption": self.image_caption,
        }

    def files(self):
        return {"image_file": (self.image_name, self.image_bytes)}


@dataclass
class KnowledgeItem:
    """
    A KnowledgeItem is a set of elements (titles, texts, urls, etc.) that are used for Retrieval Augmented Generation (RAG).
    """
    title: str
    content: str
    knowledge_base: Optional[str] = None
    data_source: Optional[str] = None
    section: Optional[str] = None
    url: Optional[str] = None
    page_number: Optional[int] = 0
    metadata: Optional[dict] = None
    images: Optional[List[KnowledgeItemImage]] = None
    id: Any = None
    created_date: Any = None
    updated_date: Any = None
    role: Any = None
    message: Any = None

    def dict(self):
        return {
            "title": self.title,
            "content": self.content,
            "knowledge_base": self.knowledge_base,
            "data_source": self.data_source,
            "section": self.section,
            "url": self.url,
            "page_number": self.page_number,
            "metadata": self.metadata,
        }


@dataclass
class DataSource:
    kb_id: str
    ds_id: Optional[str]
    task_id: Optional[str]
    csv: Optional[str]
    pdf: Optional[str]
    url: Optional[str]


class WSType(Enum):
    rpc = "rpc"
    ai = "ai"
    parse = "parse"


@dataclass
class CacheConfig:
    """
    Configuration for caching LLM responses. For now it's only needed for Gemini.
    :param name: The name of the cache.
    :param ttl: The time to live for the cache in seconds.
    """
    name: str
    ttl: int = 3600  # Default TTL of 1 hour in seconds



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, List

if TYPE_CHECKING:
    from chatfaq_sdk import ChatFAQSDK


class StatusSerializer:
    @staticmethod
    def serialize(self) -> dict:
        raise NotImplementedError

    @staticmethod
    async def deserialize(self, status) -> StatusSerializer:
        raise NotImplementedError


class FSMDefinition:
    """
    Representation of the entire FSM, this class will generate the DSM Definition sent to the ChatFAQ's back-end server
    and will do all the validation of the states/transitions and will register all the RPCs
    """

    def __init__(
        self,
        states: List[State] = [],
        transitions: List[Transition] = [],
        status_class: StatusSerializer = None,
    ):
        """

        Parameters
        ----------
        states: list of State
            All the states conforming the FSM
        transitions: list of Transition
            All the transitions conforming the FSM
        status_class: list of function
            The class for serializing and deserializing de state of the FSM in case of having one, it should implement
            StatusSerializer
        """
        self.states = states
        self.transitions = transitions
        self.status_class = status_class

    def register_rpcs(self, chatfaq_sdk: ChatFAQSDK):
        for state in self.states:
            for rpc in state.events:
                chatfaq_sdk.rpc(rpc.__name__)(rpc)

        for transition in self.transitions:
            for rpc in transition.conditions:
                chatfaq_sdk.rpc(rpc.__name__)(rpc)
            for rpc in transition.unless:
                chatfaq_sdk.rpc(rpc.__name__)(rpc)

    def to_dict_repr(self):
        return {
            "states": [item.to_dict_repr() for item in self.states],
            "transitions": [item.to_dict_repr() for item in self.transitions],
        }


class State:
    """
    The state of the FSM
    """

    def __init__(self, name: str, events: List[Callable] = [], initial: bool = False):
        """

        Parameters
        ----------
        name: str
            Descriptive name of the state
        events: list of function
            List of functions they need to be called when entering this state
        initial: bool
            If initial = True then this state will be the first entering when initializing the FSM

        """
        self.initial = initial
        self.name = name
        self.events = events

    def to_dict_repr(self):
        return {
            "initial": self.initial,
            "name": self.name,
            "events": [f.__name__ for f in self.events],
        }


class Transition:
    """
    The transition between 2 states of the FSM
    """

    def __init__(
        self,
        dest: State,
        source: State = None,
        conditions: List[Callable] = [],
        unless: List[Callable] = [],
    ):
        """

        Parameters
        ----------
        dest: State
            Mandatory, the target state if all the conditions/unless pass
        source: State
            The state where the FSM is coming from in order to execute his transition, in case 'source' is not set then
            the 'dest' state on this transition would be considered 'ubiquitous state', meaning it can be accesses from
            any other state
        conditions: function
            The functions that has to pass in order to satisfy the transition
        unless: function
            The functions that doesn't have to pass in order to satisfy the transition
        """
        self.source = source
        self.dest = dest
        self.conditions = conditions
        self.unless = unless

    def to_dict_repr(self):
        json = {
            "dest": self.dest.name,
            "unless": [f.__name__ for f in self.unless],
            "conditions": [f.__name__ for f in self.conditions],
        }
        if self.source:
            json["source"] = self.source.name
        return json
